package com.sun.imageio.plugins.jpeg;

import javax.imageio.ImageReader;
import javax.imageio.event.IIOReadProgressListener;
import javax.imageio.IIOImage;
import javax.imageio.ImageTypeSpecifier;
import javax.imageio.stream.MemoryCacheImageOutputStream;
import java.io.ByteArrayOutputStream;
import java.awt.image.Raster;
import java.awt.image.ComponentColorModel;
import java.awt.image.DataBufferByte;
import java.awt.color.ICC_Profile;
import java.awt.color.ColorSpace;
import java.awt.image.ColorModel;
import java.awt.image.IndexColorModel;
import java.util.List;
import javax.imageio.stream.ImageOutputStream;
import java.awt.image.BufferedImage;
import javax.imageio.stream.ImageInputStream;
import org.w3c.dom.NodeList;
import org.w3c.dom.NamedNodeMap;
import javax.imageio.metadata.IIOMetadataNode;
import java.awt.color.ICC_ColorSpace;
import javax.imageio.IIOException;
import java.util.Iterator;
import javax.imageio.metadata.IIOInvalidTreeException;
import org.w3c.dom.Node;
import java.io.IOException;
import java.util.ArrayList;

class JFIFMarkerSegment extends MarkerSegment
{
    int majorVersion;
    int minorVersion;
    int resUnits;
    int Xdensity;
    int Ydensity;
    int thumbWidth;
    int thumbHeight;
    JFIFMarkerSegment.JFIFThumbRGB thumb;
    ArrayList extSegments;
    JFIFMarkerSegment.ICCMarkerSegment iccSegment;
    private static final int THUMB_JPEG = 16;
    private static final int THUMB_PALETTE = 17;
    private static final int THUMB_UNASSIGNED = 18;
    private static final int THUMB_RGB = 19;
    private static final int DATA_SIZE = 14;
    private static final int ID_SIZE = 5;
    private final int MAX_THUMB_WIDTH = 255;
    private final int MAX_THUMB_HEIGHT = 255;
    private final boolean debug = false;
    private boolean inICC;
    private JFIFMarkerSegment.ICCMarkerSegment tempICCSegment;
    
    JFIFMarkerSegment() {
        super(224);
        this.thumb = null;
        this.extSegments = new ArrayList();
        this.iccSegment = null;
        this.inICC = false;
        this.tempICCSegment = null;
        this.majorVersion = 1;
        this.minorVersion = 2;
        this.resUnits = 0;
        this.Xdensity = 1;
        this.Ydensity = 1;
        this.thumbWidth = 0;
        this.thumbHeight = 0;
    }
    
    JFIFMarkerSegment(JPEGBuffer jpegBuffer) throws IOException {
        super(jpegBuffer);
        this.thumb = null;
        this.extSegments = new ArrayList();
        this.iccSegment = null;
        this.inICC = false;
        this.tempICCSegment = null;
        jpegBuffer.bufPtr += 5;
        this.majorVersion = jpegBuffer.buf[jpegBuffer.bufPtr++];
        this.minorVersion = jpegBuffer.buf[jpegBuffer.bufPtr++];
        this.resUnits = jpegBuffer.buf[jpegBuffer.bufPtr++];
        this.Xdensity = (jpegBuffer.buf[jpegBuffer.bufPtr++] & 0xFF) << 8;
        this.Xdensity |= (jpegBuffer.buf[jpegBuffer.bufPtr++] & 0xFF);
        this.Ydensity = (jpegBuffer.buf[jpegBuffer.bufPtr++] & 0xFF) << 8;
        this.Ydensity |= (jpegBuffer.buf[jpegBuffer.bufPtr++] & 0xFF);
        this.thumbWidth = (jpegBuffer.buf[jpegBuffer.bufPtr++] & 0xFF);
        this.thumbHeight = (jpegBuffer.buf[jpegBuffer.bufPtr++] & 0xFF);
        jpegBuffer.bufAvail -= 14;
        if (this.thumbWidth > 0) {
            this.thumb = new JFIFMarkerSegment.JFIFThumbRGB(jpegBuffer, this.thumbWidth, this.thumbHeight);
        }
    }
    
    JFIFMarkerSegment(Node node) throws IIOInvalidTreeException {
        this();
        this.updateFromNativeNode(node, true);
    }
    
    @Override
    protected Object clone() {
        JFIFMarkerSegment jfifMarkerSegment = (JFIFMarkerSegment)super.clone();
        if (!this.extSegments.isEmpty()) {
            jfifMarkerSegment.extSegments = new ArrayList();
            Iterator<Object> iterator = this.extSegments.iterator();
            while (iterator.hasNext()) {
                jfifMarkerSegment.extSegments.add(((JFIFMarkerSegment.JFIFExtensionMarkerSegment)iterator.next()).clone());
            }
        }
        if (this.iccSegment != null) {
            jfifMarkerSegment.iccSegment = (JFIFMarkerSegment.ICCMarkerSegment)this.iccSegment.clone();
            return jfifMarkerSegment;
        }
        return jfifMarkerSegment;
    }
    
    void addJFXX(JPEGBuffer jpegBuffer, JPEGImageReader jpegImageReader) throws IOException {
        this.extSegments.add(new JFIFMarkerSegment.JFIFExtensionMarkerSegment(jpegBuffer, jpegImageReader));
    }
    
    void addICC(JPEGBuffer jpegBuffer) throws IOException {
        if (!this.inICC) {
            if (this.iccSegment != null) {
                throw new IIOException("> 1 ICC APP2 Marker Segment not supported");
            }
            this.tempICCSegment = new JFIFMarkerSegment.ICCMarkerSegment(jpegBuffer);
            if (!this.inICC) {
                this.iccSegment = this.tempICCSegment;
                this.tempICCSegment = null;
            }
        }
        else if (this.tempICCSegment.addData(jpegBuffer)) {
            this.iccSegment = this.tempICCSegment;
            this.tempICCSegment = null;
        }
    }
    
    void addICC(ICC_ColorSpace icc_ColorSpace) throws IOException {
        if (this.iccSegment != null) {
            throw new IIOException("> 1 ICC APP2 Marker Segment not supported");
        }
        this.iccSegment = new JFIFMarkerSegment.ICCMarkerSegment(icc_ColorSpace);
        return;
    }
    
    @Override
    IIOMetadataNode getNativeNode() {
        IIOMetadataNode iioMetadataNode = new IIOMetadataNode("app0JFIF");
        iioMetadataNode.setAttribute("majorVersion", Integer.toString(this.majorVersion));
        iioMetadataNode.setAttribute("minorVersion", Integer.toString(this.minorVersion));
        iioMetadataNode.setAttribute("resUnits", Integer.toString(this.resUnits));
        iioMetadataNode.setAttribute("Xdensity", Integer.toString(this.Xdensity));
        iioMetadataNode.setAttribute("Ydensity", Integer.toString(this.Ydensity));
        iioMetadataNode.setAttribute("thumbWidth", Integer.toString(this.thumbWidth));
        iioMetadataNode.setAttribute("thumbHeight", Integer.toString(this.thumbHeight));
        if (!this.extSegments.isEmpty()) {
            IIOMetadataNode newChild = new IIOMetadataNode("JFXX");
            iioMetadataNode.appendChild(newChild);
            Iterator<Object> iterator = this.extSegments.iterator();
            while (iterator.hasNext()) {
                newChild.appendChild(((JFIFMarkerSegment.JFIFExtensionMarkerSegment)iterator.next()).getNativeNode());
            }
        }
        if (this.iccSegment != null) {
            iioMetadataNode.appendChild(this.iccSegment.getNativeNode());
            return iioMetadataNode;
        }
        return iioMetadataNode;
    }
    
    void updateFromNativeNode(Node offendingNode, boolean b) throws IIOInvalidTreeException {
        NamedNodeMap attributes = offendingNode.getAttributes();
        if (attributes.getLength() > 0) {
            int attributeValue = MarkerSegment.getAttributeValue(offendingNode, attributes, "majorVersion", 0, 255, false);
            this.majorVersion = ((attributeValue != -1) ? attributeValue : this.majorVersion);
            int attributeValue2 = MarkerSegment.getAttributeValue(offendingNode, attributes, "minorVersion", 0, 255, false);
            this.minorVersion = ((attributeValue2 != -1) ? attributeValue2 : this.minorVersion);
            int attributeValue3 = MarkerSegment.getAttributeValue(offendingNode, attributes, "resUnits", 0, 2, false);
            this.resUnits = ((attributeValue3 != -1) ? attributeValue3 : this.resUnits);
            int attributeValue4 = MarkerSegment.getAttributeValue(offendingNode, attributes, "Xdensity", 1, 65535, false);
            this.Xdensity = ((attributeValue4 != -1) ? attributeValue4 : this.Xdensity);
            int attributeValue5 = MarkerSegment.getAttributeValue(offendingNode, attributes, "Ydensity", 1, 65535, false);
            this.Ydensity = ((attributeValue5 != -1) ? attributeValue5 : this.Ydensity);
            int attributeValue6 = MarkerSegment.getAttributeValue(offendingNode, attributes, "thumbWidth", 0, 255, false);
            this.thumbWidth = ((attributeValue6 != -1) ? attributeValue6 : this.thumbWidth);
            int attributeValue7 = MarkerSegment.getAttributeValue(offendingNode, attributes, "thumbHeight", 0, 255, false);
            this.thumbHeight = ((attributeValue7 != -1) ? attributeValue7 : this.thumbHeight);
        }
        if (offendingNode.hasChildNodes()) {
            NodeList childNodes = offendingNode.getChildNodes();
            int length = childNodes.getLength();
            if (length > 2) {
                throw new IIOInvalidTreeException("app0JFIF node cannot have > 2 children", offendingNode);
            }
            for (int i = 0; i < length; ++i) {
                Node item = childNodes.item(i);
                String nodeName = item.getNodeName();
                if (nodeName.equals("JFXX")) {
                    if (!this.extSegments.isEmpty() && b) {
                        throw new IIOInvalidTreeException("app0JFIF node cannot have > 1 JFXX node", offendingNode);
                    }
                    NodeList childNodes2 = item.getChildNodes();
                    for (int length2 = childNodes2.getLength(), j = 0; j < length2; ++j) {
                        this.extSegments.add(new JFIFMarkerSegment.JFIFExtensionMarkerSegment(childNodes2.item(j)));
                    }
                }
                if (nodeName.equals("app2ICC")) {
                    if (this.iccSegment != null && b) {
                        throw new IIOInvalidTreeException("> 1 ICC APP2 Marker Segment not supported", offendingNode);
                    }
                    this.iccSegment = new JFIFMarkerSegment.ICCMarkerSegment(item);
                }
            }
        }
    }
    
    int getThumbnailWidth(int index) {
        if (this.thumb != null) {
            if (index == 0) {
                return this.thumb.getWidth();
            }
            --index;
        }
        return ((JFIFMarkerSegment.JFIFExtensionMarkerSegment)this.extSegments.get(index)).thumb.getWidth();
    }
    
    int getThumbnailHeight(int index) {
        if (this.thumb != null) {
            if (index == 0) {
                return this.thumb.getHeight();
            }
            --index;
        }
        return ((JFIFMarkerSegment.JFIFExtensionMarkerSegment)this.extSegments.get(index)).thumb.getHeight();
    }
    
    BufferedImage getThumbnail(ImageInputStream imageInputStream, int index, JPEGImageReader jpegImageReader) throws IOException {
        jpegImageReader.thumbnailStarted(index);
        BufferedImage bufferedImage;
        if (this.thumb != null && index == 0) {
            bufferedImage = this.thumb.getThumbnail(imageInputStream, jpegImageReader);
        }
        else {
            if (this.thumb != null) {
                --index;
            }
            bufferedImage = ((JFIFMarkerSegment.JFIFExtensionMarkerSegment)this.extSegments.get(index)).thumb.getThumbnail(imageInputStream, jpegImageReader);
        }
        jpegImageReader.thumbnailComplete();
        return bufferedImage;
    }
    
    void write(ImageOutputStream imageOutputStream, JPEGImageWriter jpegImageWriter) throws IOException {
        this.write(imageOutputStream, null, jpegImageWriter);
    }
    
    void write(ImageOutputStream imageOutputStream, BufferedImage bufferedImage, JPEGImageWriter jpegImageWriter) throws IOException {
        int min = 0;
        int min2 = 0;
        int length = 0;
        int[] pixels = null;
        if (bufferedImage != null) {
            int width = bufferedImage.getWidth();
            int height = bufferedImage.getHeight();
            if (width > 255 || height > 255) {
                jpegImageWriter.warningOccurred(12);
            }
            min = Math.min(width, 255);
            min2 = Math.min(height, 255);
            pixels = bufferedImage.getRaster().getPixels(0, 0, min, min2, (int[])null);
            length = pixels.length;
        }
        this.length = 16 + length;
        this.writeTag(imageOutputStream);
        imageOutputStream.write(new byte[] { 74, 70, 73, 70, 0 });
        imageOutputStream.write(this.majorVersion);
        imageOutputStream.write(this.minorVersion);
        imageOutputStream.write(this.resUnits);
        MarkerSegment.write2bytes(imageOutputStream, this.Xdensity);
        MarkerSegment.write2bytes(imageOutputStream, this.Ydensity);
        imageOutputStream.write(min);
        imageOutputStream.write(min2);
        if (pixels != null) {
            jpegImageWriter.thumbnailStarted(0);
            this.writeThumbnailData(imageOutputStream, pixels, jpegImageWriter);
            jpegImageWriter.thumbnailComplete();
        }
    }
    
    void writeThumbnailData(ImageOutputStream imageOutputStream, int[] array, JPEGImageWriter jpegImageWriter) throws IOException {
        int n = array.length / 20;
        if (n == 0) {
            n = 1;
        }
        for (int i = 0; i < array.length; ++i) {
            imageOutputStream.write(array[i]);
            if (i > n && i % n == 0) {
                jpegImageWriter.thumbnailProgress((float)i * 100.0f / (float)array.length);
            }
        }
    }
    
    void writeWithThumbs(ImageOutputStream imageOutputStream, List list, JPEGImageWriter jpegImageWriter) throws IOException {
        if (list != null) {
            JFIFMarkerSegment.JFIFExtensionMarkerSegment jfifExtensionMarkerSegment = null;
            if (list.size() == 1) {
                if (!this.extSegments.isEmpty()) {
                    jfifExtensionMarkerSegment = (JFIFMarkerSegment.JFIFExtensionMarkerSegment)this.extSegments.get(0);
                }
                this.writeThumb(imageOutputStream, (BufferedImage)list.get(0), jfifExtensionMarkerSegment, 0, true, jpegImageWriter);
            }
            else {
                this.write(imageOutputStream, jpegImageWriter);
                for (int i = 0; i < list.size(); ++i) {
                    JFIFMarkerSegment.JFIFExtensionMarkerSegment jfifExtensionMarkerSegment2 = null;
                    if (i < this.extSegments.size()) {
                        jfifExtensionMarkerSegment2 = (JFIFMarkerSegment.JFIFExtensionMarkerSegment)this.extSegments.get(i);
                    }
                    this.writeThumb(imageOutputStream, (BufferedImage)list.get(i), jfifExtensionMarkerSegment2, i, false, jpegImageWriter);
                }
            }
        }
        else {
            this.write(imageOutputStream, jpegImageWriter);
        }
    }
    
    private void writeThumb(ImageOutputStream imageOutputStream, BufferedImage bufferedImage, JFIFMarkerSegment.JFIFExtensionMarkerSegment jfifExtensionMarkerSegment, int n, boolean b, JPEGImageWriter jpegImageWriter) throws IOException {
        ColorModel colorModel = bufferedImage.getColorModel();
        ColorSpace colorSpace = colorModel.getColorSpace();
        if ((colorModel instanceof IndexColorModel)) {
            if (b) {
                this.write(imageOutputStream, jpegImageWriter);
            }
            if (jfifExtensionMarkerSegment == null || jfifExtensionMarkerSegment.code == 17) {
                this.writeJFXXSegment(n, bufferedImage, imageOutputStream, jpegImageWriter);
            }
            else {
                jfifExtensionMarkerSegment.setThumbnail(((IndexColorModel)colorModel).convertToIntDiscrete(bufferedImage.getRaster(), false));
                jpegImageWriter.thumbnailStarted(n);
                jfifExtensionMarkerSegment.write(imageOutputStream, jpegImageWriter);
                jpegImageWriter.thumbnailComplete();
            }
        }
        else if (colorSpace.getType() == 5) {
            if (jfifExtensionMarkerSegment == null) {
                if (b) {
                    this.write(imageOutputStream, bufferedImage, jpegImageWriter);
                }
                else {
                    this.writeJFXXSegment(n, bufferedImage, imageOutputStream, jpegImageWriter);
                }
            }
            else {
                if (b) {
                    this.write(imageOutputStream, jpegImageWriter);
                }
                if (jfifExtensionMarkerSegment.code == 17) {
                    this.writeJFXXSegment(n, bufferedImage, imageOutputStream, jpegImageWriter);
                    jpegImageWriter.warningOccurred(14);
                }
                else {
                    jfifExtensionMarkerSegment.setThumbnail(bufferedImage);
                    jpegImageWriter.thumbnailStarted(n);
                    jfifExtensionMarkerSegment.write(imageOutputStream, jpegImageWriter);
                    jpegImageWriter.thumbnailComplete();
                }
            }
        }
        else if (colorSpace.getType() == 6) {
            if (jfifExtensionMarkerSegment == null) {
                if (b) {
                    this.write(imageOutputStream, expandGrayThumb(bufferedImage), jpegImageWriter);
                }
                else {
                    this.writeJFXXSegment(n, bufferedImage, imageOutputStream, jpegImageWriter);
                }
            }
            else {
                if (b) {
                    this.write(imageOutputStream, jpegImageWriter);
                }
                if (jfifExtensionMarkerSegment.code == 19) {
                    this.writeJFXXSegment(n, expandGrayThumb(bufferedImage), imageOutputStream, jpegImageWriter);
                }
                else if (jfifExtensionMarkerSegment.code == 16) {
                    jfifExtensionMarkerSegment.setThumbnail(bufferedImage);
                    jpegImageWriter.thumbnailStarted(n);
                    jfifExtensionMarkerSegment.write(imageOutputStream, jpegImageWriter);
                    jpegImageWriter.thumbnailComplete();
                }
                else if (jfifExtensionMarkerSegment.code == 17) {
                    this.writeJFXXSegment(n, bufferedImage, imageOutputStream, jpegImageWriter);
                    jpegImageWriter.warningOccurred(15);
                }
            }
        }
        else {
            jpegImageWriter.warningOccurred(9);
        }
    }
    
    private void writeJFXXSegment(int n, BufferedImage bufferedImage, ImageOutputStream imageOutputStream, JPEGImageWriter jpegImageWriter) throws IOException {
        JFIFMarkerSegment.JFIFExtensionMarkerSegment jfifExtensionMarkerSegment;
        try {
            jfifExtensionMarkerSegment = new JFIFMarkerSegment.JFIFExtensionMarkerSegment(bufferedImage);
        }
        catch (JFIFMarkerSegment.IllegalThumbException ex) {
            jpegImageWriter.warningOccurred(9);
            return;
        }
        jpegImageWriter.thumbnailStarted(n);
        jfifExtensionMarkerSegment.write(imageOutputStream, jpegImageWriter);
        jpegImageWriter.thumbnailComplete();
    }
    
    private static BufferedImage expandGrayThumb(BufferedImage bufferedImage) {
        BufferedImage bufferedImage2 = new BufferedImage(bufferedImage.getWidth(), bufferedImage.getHeight(), 1);
        bufferedImage2.getGraphics().drawImage(bufferedImage, 0, 0, null);
        return bufferedImage2;
    }
    
    static void writeDefaultJFIF(ImageOutputStream imageOutputStream, List list, ICC_Profile icc_Profile, JPEGImageWriter jpegImageWriter) throws IOException {
        new JFIFMarkerSegment().writeWithThumbs(imageOutputStream, list, jpegImageWriter);
        if (icc_Profile != null) {
            writeICC(icc_Profile, imageOutputStream);
        }
    }
    
    @Override
    void print() {
        this.printTag("JFIF");
        System.out.print("Version ");
        System.out.print(this.majorVersion);
        System.out.println(".0" + Integer.toString(this.minorVersion));
        System.out.print("Resolution units: ");
        System.out.println(this.resUnits);
        System.out.print("X density: ");
        System.out.println(this.Xdensity);
        System.out.print("Y density: ");
        System.out.println(this.Ydensity);
        System.out.print("Thumbnail Width: ");
        System.out.println(this.thumbWidth);
        System.out.print("Thumbnail Height: ");
        System.out.println(this.thumbHeight);
        if (!this.extSegments.isEmpty()) {
            Iterator<Object> iterator = this.extSegments.iterator();
            while (iterator.hasNext()) {
                ((JFIFMarkerSegment.JFIFExtensionMarkerSegment)iterator.next()).print();
            }
        }
        if (this.iccSegment != null) {
            this.iccSegment.print();
        }
    }
    
    static void writeICC(ICC_Profile icc_Profile, ImageOutputStream imageOutputStream) throws IOException {
        int n = 2;
        int n2 = "ICC_PROFILE".length() + 1;
        int n3 = 2;
        int b = 65535 - n - n2 - n3;
        byte[] data = icc_Profile.getData();
        int n4 = data.length / b;
        if (data.length % b != 0) {
            ++n4;
        }
        int n5 = 1;
        int min;
        for (int n6 = 0, i = 0; i < n4; n6 += min, ++i) {
            min = Math.min(data.length - n6, b);
            int n7 = min + n3 + n2 + n;
            imageOutputStream.write(255);
            imageOutputStream.write(226);
            MarkerSegment.write2bytes(imageOutputStream, n7);
            imageOutputStream.write("ICC_PROFILE".getBytes("US-ASCII"));
            imageOutputStream.write(0);
            imageOutputStream.write(n5++);
            imageOutputStream.write(n4);
            imageOutputStream.write(data, n6, min);
        }
    }
    
    private class IllegalThumbException extends Exception
    {
    }
    
    class JFIFExtensionMarkerSegment extends MarkerSegment
    {
        int code;
        JFIFMarkerSegment.JFIFThumb thumb;
        private static final int DATA_SIZE = 6;
        private static final int ID_SIZE = 5;
        
        JFIFExtensionMarkerSegment(JPEGBuffer jpegBuffer, JPEGImageReader jpegImageReader) throws IOException {
            super(jpegBuffer);
            jpegBuffer.bufPtr += 5;
            this.code = (jpegBuffer.buf[jpegBuffer.bufPtr++] & 0xFF);
            jpegBuffer.bufAvail -= 6;
            if (this.code == 16) {
                this.thumb = new JFIFMarkerSegment.JFIFThumbJPEG(jpegBuffer, this.length, jpegImageReader);
            }
            else {
                jpegBuffer.loadBuf(2);
                int n = jpegBuffer.buf[jpegBuffer.bufPtr++] & 0xFF;
                int n2 = jpegBuffer.buf[jpegBuffer.bufPtr++] & 0xFF;
                jpegBuffer.bufAvail -= 2;
                if (this.code == 17) {
                    this.thumb = new JFIFMarkerSegment.JFIFThumbPalette(jpegBuffer, n, n2);
                }
                else {
                    this.thumb = new JFIFMarkerSegment.JFIFThumbRGB(jpegBuffer, n, n2);
                }
            }
        }
        
        JFIFExtensionMarkerSegment(Node offendingNode) throws IIOInvalidTreeException {
            super(224);
            NamedNodeMap attributes = offendingNode.getAttributes();
            if (attributes.getLength() > 0) {
                this.code = MarkerSegment.getAttributeValue(offendingNode, attributes, "extensionCode", 16, 19, false);
                if (this.code == 18) {
                    throw new IIOInvalidTreeException("invalid extensionCode attribute value", offendingNode);
                }
            }
            else {
                this.code = 18;
            }
            if (offendingNode.getChildNodes().getLength() != 1) {
                throw new IIOInvalidTreeException("app0JFXX node must have exactly 1 child", offendingNode);
            }
            Node firstChild = offendingNode.getFirstChild();
            String nodeName = firstChild.getNodeName();
            if (nodeName.equals("JFIFthumbJPEG")) {
                if (this.code == 18) {
                    this.code = 16;
                }
                this.thumb = new JFIFMarkerSegment.JFIFThumbJPEG(firstChild);
            }
            else if (nodeName.equals("JFIFthumbPalette")) {
                if (this.code == 18) {
                    this.code = 17;
                }
                this.thumb = new JFIFMarkerSegment.JFIFThumbPalette(firstChild);
            }
            else if (nodeName.equals("JFIFthumbRGB")) {
                if (this.code == 18) {
                    this.code = 19;
                }
                this.thumb = new JFIFMarkerSegment.JFIFThumbRGB(firstChild);
            }
            else {
                throw new IIOInvalidTreeException("unrecognized app0JFXX child node", offendingNode);
            }
            return;
        }
        
        JFIFExtensionMarkerSegment(BufferedImage bufferedImage) throws JFIFMarkerSegment.IllegalThumbException {
            super(224);
            ColorModel colorModel = bufferedImage.getColorModel();
            int type = colorModel.getColorSpace().getType();
            if (colorModel.hasAlpha()) {
                throw new JFIFMarkerSegment.IllegalThumbException();
            }
            if ((colorModel instanceof IndexColorModel)) {
                this.code = 17;
                this.thumb = new JFIFMarkerSegment.JFIFThumbPalette(bufferedImage);
            }
            else if (type == 5) {
                this.code = 19;
                this.thumb = new JFIFMarkerSegment.JFIFThumbRGB(bufferedImage);
            }
            else if (type == 6) {
                this.code = 16;
                this.thumb = new JFIFMarkerSegment.JFIFThumbJPEG(bufferedImage);
            }
            else {
                throw new JFIFMarkerSegment.IllegalThumbException();
            }
            return;
        }
        
        void setThumbnail(BufferedImage bufferedImage) {
            try {
                switch (this.code) {
                    case 17: {
                        this.thumb = new JFIFMarkerSegment.JFIFThumbPalette(bufferedImage);
                        break;
                    }
                    case 19: {
                        this.thumb = new JFIFMarkerSegment.JFIFThumbRGB(bufferedImage);
                        break;
                    }
                    case 16: {
                        this.thumb = new JFIFMarkerSegment.JFIFThumbJPEG(bufferedImage);
                        break;
                    }
                }
            }
            catch (JFIFMarkerSegment.IllegalThumbException cause) {
                throw new InternalError("Illegal thumb in setThumbnail!", cause);
            }
        }
        
        @Override
        protected Object clone() {
            JFIFMarkerSegment.JFIFExtensionMarkerSegment jfifExtensionMarkerSegment = (JFIFMarkerSegment.JFIFExtensionMarkerSegment)super.clone();
            if (this.thumb != null) {
                jfifExtensionMarkerSegment.thumb = (JFIFMarkerSegment.JFIFThumb)this.thumb.clone();
                return jfifExtensionMarkerSegment;
            }
            return jfifExtensionMarkerSegment;
        }
        
        @Override
        IIOMetadataNode getNativeNode() {
            IIOMetadataNode iioMetadataNode = new IIOMetadataNode("app0JFXX");
            iioMetadataNode.setAttribute("extensionCode", Integer.toString(this.code));
            iioMetadataNode.appendChild(this.thumb.getNativeNode());
            return iioMetadataNode;
        }
        
        void write(ImageOutputStream imageOutputStream, JPEGImageWriter jpegImageWriter) throws IOException {
            this.length = 8 + this.thumb.getLength();
            this.writeTag(imageOutputStream);
            imageOutputStream.write(new byte[] { 74, 70, 88, 88, 0 });
            imageOutputStream.write(this.code);
            this.thumb.write(imageOutputStream, jpegImageWriter);
        }
        
        @Override
        void print() {
            this.printTag("JFXX");
            this.thumb.print();
        }
    }
    
    abstract class JFIFThumb implements Cloneable
    {
        long streamPos;
        
        abstract int getLength();
        
        abstract int getWidth();
        
        abstract int getHeight();
        
        abstract BufferedImage getThumbnail(ImageInputStream p0, JPEGImageReader p1) throws IOException;
        
        protected JFIFThumb() {
            this.streamPos = -1L;
        }
        
        protected JFIFThumb(JPEGBuffer jpegBuffer) throws IOException {
            this.streamPos = -1L;
            this.streamPos = jpegBuffer.getStreamPosition();
        }
        
        abstract void print();
        
        abstract IIOMetadataNode getNativeNode();
        
        abstract void write(ImageOutputStream p0, JPEGImageWriter p1) throws IOException;
        
        @Override
        protected Object clone() {
            try {
                return super.clone();
            }
            catch (CloneNotSupportedException ex) {
                return null;
            }
        }
    }
    
    abstract class JFIFThumbUncompressed extends JFIFMarkerSegment.JFIFThumb
    {
        BufferedImage thumbnail;
        int thumbWidth;
        int thumbHeight;
        String name;
        
        JFIFThumbUncompressed(JPEGBuffer jpegBuffer, int thumbWidth, int thumbHeight, int n, String name) throws IOException {
            super(jpegBuffer);
            this.thumbnail = null;
            this.thumbWidth = thumbWidth;
            this.thumbHeight = thumbHeight;
            jpegBuffer.skipData(n);
            this.name = name;
        }
        
        JFIFThumbUncompressed(Node offendingNode, String s) throws IIOInvalidTreeException {
            this.thumbnail = null;
            this.thumbWidth = 0;
            this.thumbHeight = 0;
            this.name = s;
            NamedNodeMap attributes = offendingNode.getAttributes();
            int length = attributes.getLength();
            if (length > 2) {
                throw new IIOInvalidTreeException(s + " node cannot have > 2 attributes", offendingNode);
            }
            if (length != 0) {
                int attributeValue = MarkerSegment.getAttributeValue(offendingNode, attributes, "thumbWidth", 0, 255, false);
                this.thumbWidth = ((attributeValue != -1) ? attributeValue : this.thumbWidth);
                int attributeValue2 = MarkerSegment.getAttributeValue(offendingNode, attributes, "thumbHeight", 0, 255, false);
                this.thumbHeight = ((attributeValue2 != -1) ? attributeValue2 : this.thumbHeight);
            }
            return;
        }
        
        JFIFThumbUncompressed(BufferedImage thumbnail) {
            this.thumbnail = null;
            this.thumbnail = thumbnail;
            this.thumbWidth = thumbnail.getWidth();
            this.thumbHeight = thumbnail.getHeight();
            this.name = null;
        }
        
        void readByteBuffer(ImageInputStream imageInputStream, byte[] array, JPEGImageReader jpegImageReader, float n, float n2) throws IOException {
            int max = Math.max((int)((float)(array.length / 20) / n), 1);
            int i = 0;
            while (i < array.length) {
                imageInputStream.read(array, i, Math.min(max, array.length - i));
                i += max;
                float n3 = (float)i * 100.0f / (float)array.length * n + n2;
                if (n3 > 100.0f) {
                    n3 = 100.0f;
                }
                jpegImageReader.thumbnailProgress(n3);
            }
        }
        
        @Override
        int getWidth() {
            return this.thumbWidth;
        }
        
        @Override
        int getHeight() {
            return this.thumbHeight;
        }
        
        @Override
        IIOMetadataNode getNativeNode() {
            IIOMetadataNode iioMetadataNode = new IIOMetadataNode(this.name);
            iioMetadataNode.setAttribute("thumbWidth", Integer.toString(this.thumbWidth));
            iioMetadataNode.setAttribute("thumbHeight", Integer.toString(this.thumbHeight));
            return iioMetadataNode;
        }
        
        @Override
        void write(ImageOutputStream imageOutputStream, JPEGImageWriter jpegImageWriter) throws IOException {
            if (this.thumbWidth > 255 || this.thumbHeight > 255) {
                jpegImageWriter.warningOccurred(12);
            }
            this.thumbWidth = Math.min(this.thumbWidth, 255);
            this.thumbHeight = Math.min(this.thumbHeight, 255);
            imageOutputStream.write(this.thumbWidth);
            imageOutputStream.write(this.thumbHeight);
        }
        
        void writePixels(ImageOutputStream imageOutputStream, JPEGImageWriter jpegImageWriter) throws IOException {
            if (this.thumbWidth > 255 || this.thumbHeight > 255) {
                jpegImageWriter.warningOccurred(12);
            }
            this.thumbWidth = Math.min(this.thumbWidth, 255);
            this.thumbHeight = Math.min(this.thumbHeight, 255);
            JFIFMarkerSegment.this.writeThumbnailData(imageOutputStream, this.thumbnail.getRaster().getPixels(0, 0, this.thumbWidth, this.thumbHeight, (int[])null), jpegImageWriter);
        }
        
        @Override
        void print() {
            System.out.print(this.name + " width: ");
            System.out.println(this.thumbWidth);
            System.out.print(this.name + " height: ");
            System.out.println(this.thumbHeight);
        }
    }
    
    class JFIFThumbRGB extends JFIFMarkerSegment.JFIFThumbUncompressed
    {
        JFIFThumbRGB(JPEGBuffer jpegBuffer, int n, int n2) throws IOException {
            super(jpegBuffer, n, n2, n * n2 * 3, "JFIFthumbRGB");
        }
        
        JFIFThumbRGB(Node node) throws IIOInvalidTreeException {
            super(node, "JFIFthumbRGB");
        }
        
        JFIFThumbRGB(BufferedImage bufferedImage) throws JFIFMarkerSegment.IllegalThumbException {
            super(bufferedImage);
        }
        
        @Override
        int getLength() {
            return this.thumbWidth * this.thumbHeight * 3;
        }
        
        @Override
        BufferedImage getThumbnail(ImageInputStream imageInputStream, JPEGImageReader jpegImageReader) throws IOException {
            imageInputStream.mark();
            imageInputStream.seek(this.streamPos);
            DataBufferByte dataBuffer = new DataBufferByte(this.getLength());
            this.readByteBuffer(imageInputStream, dataBuffer.getData(), jpegImageReader, 1.0f, 0.0f);
            imageInputStream.reset();
            return new BufferedImage(new ComponentColorModel(JPEG.JCS.sRGB, false, false, 1, 0), Raster.createInterleavedRaster(dataBuffer, this.thumbWidth, this.thumbHeight, this.thumbWidth * 3, 3, new int[] { 0, 1, 2 }, null), false, null);
        }
        
        @Override
        void write(ImageOutputStream imageOutputStream, JPEGImageWriter jpegImageWriter) throws IOException {
            super.write(imageOutputStream, jpegImageWriter);
            this.writePixels(imageOutputStream, jpegImageWriter);
        }
    }
    
    class JFIFThumbPalette extends JFIFMarkerSegment.JFIFThumbUncompressed
    {
        private static final int PALETTE_SIZE = 768;
        
        JFIFThumbPalette(JPEGBuffer jpegBuffer, int n, int n2) throws IOException {
            super(jpegBuffer, n, n2, 768 + n * n2, "JFIFThumbPalette");
        }
        
        JFIFThumbPalette(Node node) throws IIOInvalidTreeException {
            super(node, "JFIFThumbPalette");
        }
        
        JFIFThumbPalette(BufferedImage bufferedImage) throws JFIFMarkerSegment.IllegalThumbException {
            super(bufferedImage);
            if (((IndexColorModel)this.thumbnail.getColorModel()).getMapSize() > 256) {
                throw new JFIFMarkerSegment.IllegalThumbException();
            }
            return;
        }
        
        @Override
        int getLength() {
            return this.thumbWidth * this.thumbHeight + 768;
        }
        
        @Override
        BufferedImage getThumbnail(ImageInputStream imageInputStream, JPEGImageReader jpegImageReader) throws IOException {
            imageInputStream.mark();
            imageInputStream.seek(this.streamPos);
            byte[] cmap = new byte[768];
            float n = 768.0f / (float)this.getLength();
            this.readByteBuffer(imageInputStream, cmap, jpegImageReader, n, 0.0f);
            DataBufferByte db = new DataBufferByte(this.thumbWidth * this.thumbHeight);
            this.readByteBuffer(imageInputStream, db.getData(), jpegImageReader, 1.0f - n, n);
            imageInputStream.read();
            imageInputStream.reset();
            IndexColorModel cm = new IndexColorModel(8, 256, cmap, 0, false);
            return new BufferedImage(cm, Raster.createWritableRaster(cm.createCompatibleSampleModel(this.thumbWidth, this.thumbHeight), db, null), false, null);
        }
        
        @Override
        void write(ImageOutputStream imageOutputStream, JPEGImageWriter jpegImageWriter) throws IOException {
            super.write(imageOutputStream, jpegImageWriter);
            byte[] array = new byte[768];
            IndexColorModel indexColorModel = (IndexColorModel)this.thumbnail.getColorModel();
            byte[] r = new byte[256];
            byte[] g = new byte[256];
            byte[] b = new byte[256];
            indexColorModel.getReds(r);
            indexColorModel.getGreens(g);
            indexColorModel.getBlues(b);
            for (int i = 0; i < 256; ++i) {
                array[i * 3] = r[i];
                array[i * 3 + 1] = g[i];
                array[i * 3 + 2] = b[i];
            }
            imageOutputStream.write(array);
            this.writePixels(imageOutputStream, jpegImageWriter);
        }
    }
    
    class JFIFThumbJPEG extends JFIFMarkerSegment.JFIFThumb
    {
        JPEGMetadata thumbMetadata;
        byte[] data;
        private static final int PREAMBLE_SIZE = 6;
        
        JFIFThumbJPEG(JPEGBuffer jpegBuffer, int n, JPEGImageReader jpegImageReader) throws IOException {
            super(jpegBuffer);
            this.thumbMetadata = null;
            this.data = null;
            long n2 = this.streamPos + (long)(n - 6);
            jpegBuffer.iis.seek(this.streamPos);
            this.thumbMetadata = new JPEGMetadata(false, true, jpegBuffer.iis, jpegImageReader);
            jpegBuffer.iis.seek(n2);
            jpegBuffer.bufAvail = 0;
            jpegBuffer.bufPtr = 0;
        }
        
        JFIFThumbJPEG(Node node) throws IIOInvalidTreeException {
            this.thumbMetadata = null;
            this.data = null;
            if (node.getChildNodes().getLength() > 1) {
                throw new IIOInvalidTreeException("JFIFThumbJPEG node must have 0 or 1 child", node);
            }
            Node firstChild = node.getFirstChild();
            if (firstChild != null) {
                if (!firstChild.getNodeName().equals("markerSequence")) {
                    throw new IIOInvalidTreeException("JFIFThumbJPEG child must be a markerSequence node", node);
                }
                (this.thumbMetadata = new JPEGMetadata(false, true)).setFromMarkerSequenceNode(firstChild);
            }
            return;
        }
        
        JFIFThumbJPEG(BufferedImage bufferedImage) throws JFIFMarkerSegment.IllegalThumbException {
            this.thumbMetadata = null;
            this.data = null;
            int size = 4096;
            int n = 65527;
            try {
                ByteArrayOutputStream stream = new ByteArrayOutputStream(size);
                MemoryCacheImageOutputStream output = new MemoryCacheImageOutputStream(stream);
                JPEGImageWriter jpegImageWriter = new JPEGImageWriter(null);
                jpegImageWriter.setOutput(output);
                JPEGMetadata metadata = (JPEGMetadata)jpegImageWriter.getDefaultImageMetadata(new ImageTypeSpecifier(bufferedImage), null);
                MarkerSegment markerSegment = metadata.findMarkerSegment(JFIFMarkerSegment.class, true);
                if (markerSegment == null) {
                    throw new JFIFMarkerSegment.IllegalThumbException();
                }
                metadata.markerSequence.remove(markerSegment);
                jpegImageWriter.write(new IIOImage(bufferedImage, null, metadata));
                jpegImageWriter.dispose();
                if (stream.size() > n) {
                    throw new JFIFMarkerSegment.IllegalThumbException();
                }
                this.data = stream.toByteArray();
            }
            catch (IOException ex) {
                throw new JFIFMarkerSegment.IllegalThumbException();
            }
        }
        
        @Override
        int getWidth() {
            int samplesPerLine = 0;
            SOFMarkerSegment sofMarkerSegment = (SOFMarkerSegment)this.thumbMetadata.findMarkerSegment(SOFMarkerSegment.class, true);
            if (sofMarkerSegment != null) {
                samplesPerLine = sofMarkerSegment.samplesPerLine;
            }
            return samplesPerLine;
        }
        
        @Override
        int getHeight() {
            int numLines = 0;
            SOFMarkerSegment sofMarkerSegment = (SOFMarkerSegment)this.thumbMetadata.findMarkerSegment(SOFMarkerSegment.class, true);
            if (sofMarkerSegment != null) {
                numLines = sofMarkerSegment.numLines;
            }
            return numLines;
        }
        
        @Override
        BufferedImage getThumbnail(ImageInputStream input, JPEGImageReader jpegImageReader) throws IOException {
            input.mark();
            input.seek(this.streamPos);
            JPEGImageReader jpegImageReader2 = new JPEGImageReader(null);
            jpegImageReader2.setInput(input);
            jpegImageReader2.addIIOReadProgressListener(new JFIFMarkerSegment.JFIFThumbJPEG.ThumbnailReadListener(jpegImageReader));
            BufferedImage read = jpegImageReader2.read(0, null);
            jpegImageReader2.dispose();
            input.reset();
            return read;
        }
        
        @Override
        protected Object clone() {
            JFIFMarkerSegment.JFIFThumbJPEG jfifThumbJPEG = (JFIFMarkerSegment.JFIFThumbJPEG)super.clone();
            if (this.thumbMetadata != null) {
                jfifThumbJPEG.thumbMetadata = (JPEGMetadata)this.thumbMetadata.clone();
                return jfifThumbJPEG;
            }
            return jfifThumbJPEG;
        }
        
        @Override
        IIOMetadataNode getNativeNode() {
            IIOMetadataNode iioMetadataNode = new IIOMetadataNode("JFIFthumbJPEG");
            if (this.thumbMetadata != null) {
                iioMetadataNode.appendChild(this.thumbMetadata.getNativeTree());
                return iioMetadataNode;
            }
            return iioMetadataNode;
        }
        
        @Override
        int getLength() {
            if (this.data == null) {
                return 0;
            }
            return this.data.length;
        }
        
        @Override
        void write(ImageOutputStream imageOutputStream, JPEGImageWriter jpegImageWriter) throws IOException {
            int a = this.data.length / 20;
            if (a == 0) {
                a = 1;
            }
            int i = 0;
            while (i < this.data.length) {
                imageOutputStream.write(this.data, i, Math.min(a, this.data.length - i));
                i += a;
                float n = (float)i * 100.0f / (float)this.data.length;
                if (n > 100.0f) {
                    n = 100.0f;
                }
                jpegImageWriter.thumbnailProgress(n);
            }
        }
        
        @Override
        void print() {
            System.out.println("JFIF thumbnail stored as JPEG");
        }
        
        private class ThumbnailReadListener implements IIOReadProgressListener
        {
            JPEGImageReader reader;
            
            ThumbnailReadListener(JPEGImageReader reader) {
                this.reader = null;
                this.reader = reader;
            }
            
            @Override
            public void sequenceStarted(ImageReader imageReader, int n) {
            }
            
            @Override
            public void sequenceComplete(ImageReader imageReader) {
            }
            
            @Override
            public void imageStarted(ImageReader imageReader, int n) {
            }
            
            @Override
            public void imageProgress(ImageReader imageReader, float n) {
                this.reader.thumbnailProgress(n);
            }
            
            @Override
            public void imageComplete(ImageReader imageReader) {
            }
            
            @Override
            public void thumbnailStarted(ImageReader imageReader, int n, int n2) {
            }
            
            @Override
            public void thumbnailProgress(ImageReader imageReader, float n) {
            }
            
            @Override
            public void thumbnailComplete(ImageReader imageReader) {
            }
            
            @Override
            public void readAborted(ImageReader imageReader) {
            }
        }
    }
    
    class ICCMarkerSegment extends MarkerSegment
    {
        ArrayList chunks;
        byte[] profile;
        private static final int ID_SIZE = 12;
        int chunksRead;
        int numChunks;
        
        ICCMarkerSegment(ICC_ColorSpace icc_ColorSpace) {
            super(226);
            this.chunks = null;
            this.profile = null;
            this.chunks = null;
            this.chunksRead = 0;
            this.numChunks = 0;
            this.profile = icc_ColorSpace.getProfile().getData();
        }
        
        ICCMarkerSegment(JPEGBuffer jpegBuffer) throws IOException {
            super(jpegBuffer);
            this.chunks = null;
            this.profile = null;
            jpegBuffer.bufPtr += 12;
            jpegBuffer.bufAvail -= 12;
            this.length -= 12;
            int n = jpegBuffer.buf[jpegBuffer.bufPtr] & 0xFF;
            this.numChunks = (jpegBuffer.buf[jpegBuffer.bufPtr + 1] & 0xFF);
            if (n > this.numChunks) {
                throw new IIOException("Image format Error; chunk num > num chunks");
            }
            if (this.numChunks == 1) {
                this.length -= 2;
                this.profile = new byte[this.length];
                jpegBuffer.bufPtr += 2;
                jpegBuffer.bufAvail -= 2;
                jpegBuffer.readData(this.profile);
                JFIFMarkerSegment.this.inICC = false;
            }
            else {
                byte[] e = new byte[this.length];
                this.length -= 2;
                jpegBuffer.readData(e);
                (this.chunks = new ArrayList()).add(e);
                this.chunksRead = 1;
                JFIFMarkerSegment.this.inICC = true;
            }
            return;
        }
        
        ICCMarkerSegment(Node node) throws IIOInvalidTreeException {
            super(226);
            this.chunks = null;
            this.profile = null;
            if ((node instanceof IIOMetadataNode)) {
                ICC_Profile icc_Profile = (ICC_Profile)((IIOMetadataNode)node).getUserObject();
                if (icc_Profile != null) {
                    this.profile = icc_Profile.getData();
                }
            }
        }
        
        @Override
        protected Object clone() {
            JFIFMarkerSegment.ICCMarkerSegment iccMarkerSegment = (JFIFMarkerSegment.ICCMarkerSegment)super.clone();
            if (this.profile != null) {
                iccMarkerSegment.profile = (byte[])(byte[])this.profile.clone();
                return iccMarkerSegment;
            }
            return iccMarkerSegment;
        }
        
        boolean addData(JPEGBuffer jpegBuffer) throws IOException {
            ++jpegBuffer.bufPtr;
            --jpegBuffer.bufAvail;
            int n = (jpegBuffer.buf[jpegBuffer.bufPtr++] & 0xFF) << 8 | (jpegBuffer.buf[jpegBuffer.bufPtr++] & 0xFF);
            jpegBuffer.bufAvail -= 2;
            n -= 2;
            jpegBuffer.bufPtr += 12;
            jpegBuffer.bufAvail -= 12;
            n -= 12;
            if ((jpegBuffer.buf[jpegBuffer.bufPtr] & 0xFF) > this.numChunks) {
                throw new IIOException("Image format Error; chunk num > num chunks");
            }
            if (this.numChunks != (jpegBuffer.buf[jpegBuffer.bufPtr + 1] & 0xFF)) {
                throw new IIOException("Image format Error; icc num chunks mismatch");
            }
            n -= 2;
            boolean n2 = false;
            byte[] e = new byte[n];
            jpegBuffer.readData(e);
            this.chunks.add(e);
            this.length += n;
            ++this.chunksRead;
            if (this.chunksRead < this.numChunks) {
                JFIFMarkerSegment.this.inICC = true;
            }
            else {
                this.profile = new byte[this.length];
                int n3 = 0;
                int i = 1;
                while (i <= this.numChunks) {
                    boolean n4 = false;
                    for (int j = 0; j < this.chunks.size(); ++j) {
                        byte[] array = (byte[])(byte[])this.chunks.get(j);
                        if (array[0] == i) {
                            System.arraycopy(array, 2, this.profile, n3, array.length - 2);
                            n3 += array.length - 2;
                            n4 = true;
                        }
                    }
                    if (n4 == 0) {
                        throw new IIOException("Image Format Error: Missing ICC chunk num " + i);
                    }
                    ++i;
                }
                this.chunks = null;
                this.chunksRead = 0;
                this.numChunks = 0;
                JFIFMarkerSegment.this.inICC = false;
                n2 = true;
            }
            return n2;
        }
        
        @Override
        IIOMetadataNode getNativeNode() {
            IIOMetadataNode iioMetadataNode = new IIOMetadataNode("app2ICC");
            if (this.profile != null) {
                iioMetadataNode.setUserObject(ICC_Profile.getInstance(this.profile));
                return iioMetadataNode;
            }
            return iioMetadataNode;
        }
        
        @Override
        void write(ImageOutputStream imageOutputStream) throws IOException {
        }
        
        @Override
        void print() {
            this.printTag("ICC Profile APP2");
        }
    }
}
